#include "NoDataAvailableException.hpp"
#include <SentecSensors.h>

/***************************************
 *          RS485 SENSOR READOUT
 ****************************************/
static const char *TAG = "SENTEC";

SentecSensorRS485::SentecSensorRS485(HardwareSerial *ser, byte add) {
    address = add;
    RS485 = ser;
}

SentecSensorRS485::SentecSensorRS485(HardwareSerial *ser, byte add, uint8_t serialControlPin) {
    address = add;
    RS485 = ser;
    serialCommunicationControlPin = serialControlPin;
}

void SentecSensorRS485::write(byte queryFrame[], int length) {
    // sends a message (bytes) to the sensor

    // Initialize the transmitter
    digitalWrite(serialCommunicationControlPin, HIGH);
    // Send message: request a reading from the sensor
    RS485->write(queryFrame, length);
    RS485->flush();
    // Initialize the receiver
    digitalWrite(serialCommunicationControlPin, LOW);
}

String SentecSensorRS485::getValueStr(float value) {
    if (valid) {
        return String(value, 1);
    } else {
        return String("null");
    }
}

String SentecSensorRS485::getValueStr(int value) {
    if (valid) {
        return String(value);
    } else {
        return String("null");
    }
}

ErrorType SentecSensorRS485::queryAddress() {
    // request the address of the sensor with ONLY ONE SENSOR ON THE BUS

    byte tmp_addr = address; // store the address in a temporary byte
    address = 0xFF;          // change the address to FF (0) for address check
    ErrorType tmp = readRegister(word(0x07, 0xD0), 2);
    address = tmp_addr; // set the original address back
    return tmp;
}

ErrorType SentecSensorRS485::readRegister(int registerStartAddress) {
    return readRegister(registerStartAddress, 1);
}

ErrorType SentecSensorRS485::readRegister(int registerStartAddress, int registerLength) {
    // function code 0x03: get data measured by the sensor
    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x03;
    // register start address
    query[2] = registerStartAddress >> 8;
    query[3] = registerStartAddress & 0xFF;
    // register length
    query[4] = registerLength >> 8;
    query[5] = registerLength & 0xFF;
    // calculate last two bytes (CRC check)
    calculateCRC(query, sizeof(query) - 2);
    // #   Serial.print("Query (get data): 0x");                    #Print bytes
    // #   printBytes(query, 8);
    //  write the data request to the modbus line
    write(query, sizeof(query));
    // get response from sensor
    return getResponse();
}

ErrorType SentecSensorRS485::writeRegister(int registerAddress, int value) {
    // function code 0x06: change sensor settings
    // e.g. a new address, reset rainfal data...

    byte query[8];
    query[0] = address;
    // function code
    query[1] = 0x06;
    // register start address
    query[2] = registerAddress >> 8;
    query[3] = registerAddress & 0xFF;
    // register length
    query[4] = value >> 8;
    query[5] = value & 0xFF;
    calculateCRC(query, sizeof(query) - 2);
    ESP_LOGD(TAG, "Query (settings):    ");
    printBytes(query, 8);
    write(query, sizeof(query));
    return getResponse();
}

ErrorType SentecSensorRS485::setAddress(byte add) {
    // change the address of a sensor
    ErrorType tmp = writeRegister(word(0x07, 0xD0), add);
    if (tmp == ErrorType::DATA_OK)
        address = add;
    // TODO check response: matches the sent message exactly
    return tmp;
}

void SentecSensorRS485::resetAnswerFrame() {
    for (int i = 0; i < 10; i++) {
        answerFrame[i] = 0;
    }
}

ErrorType SentecSensorRS485::getResponse() {
    // reads the response of a sensor
    ErrorType returnCode = ErrorType::DATA_OK;
    valid = true;
    int idx = 0;
    int byteReceived;
    // usual response length: changed in the while loop to match the response,
    //   changed only when reading data (then it's 7 or 9  bytes, sensor dpendent)
    int responseLength = 8;
    // reading an answer takes up to 39 milliseconds for 2 byte readRegister
    const int timeout = 200;
    const int retries = 1; // #editet to q
    size_t tries = 1;
    // it doesn't seem to help to request multiple times if first time goes wrong
    for (tries; tries <= retries; tries++) {
        // if we lose connection with the sensor, we get an array of zeros back
        resetAnswerFrame();

        unsigned long time = millis();
        while (idx < responseLength && (millis() - time) < timeout) {
            if (RS485->available()) {
                byteReceived = RS485->read();
                // Serial.println(byteReceived, HEX);
                // check for first byte. It has to be the device address unless for broadcasts with address = 0xFF
                if (idx == 0 && address != 0xFF && byteReceived != address) {
                    ESP_LOGE(TAG,
                             "Invalid byte. First byte needs to be address 0x%02X but got 0x%02Xinstead");
                } else {
                    answerFrame[idx] = byteReceived;
                    // for reading register: third received byte is data length, read number of bytes accordingly
                    if (idx == 2 && answerFrame[1] == 0x03) {
                        // 5 bytes for address, function code, data length, CRC_H, CRC_L
                        responseLength = 5 + byteReceived;
                    }
                    idx++;
                }
            }
        }
        ESP_LOGD(TAG,
                 "-----------------------------------------------------------------------------------------");
        ESP_LOGD(TAG, "Response: 0x%s", formatBytes(answerFrame, responseLength).c_str());
        // ESP_LOGD(TAG, "Tries: %d", tries);
        // ESP_LOGD(TAG, "Bytes received: %d", idx);
        if (answerFrame[0] == 0) {
            ESP_LOGE(TAG, "Check sensor connection. First byte is 0x00.");
            valid = false;
            returnCode = ErrorType::SENSOR_NOT_CONNECTED;
        } else if (idx < responseLength) {
            ESP_LOGE(TAG, "Response too short: %d bytes < %d bytes. Unfinished transmission.", idx,
                     responseLength);
            valid = false;
            returnCode = ErrorType::CONNECTION_ENDED_PREMATURELY;
        }
        word crc_received = word(answerFrame[responseLength - 2], answerFrame[responseLength - 1]);
        word crc = calculateCRC(answerFrame, responseLength - 2);
        if (valid && crc_received != word(crc)) {
            ESP_LOGE(TAG, "CRC wrong: Expected 0x%s got 0x%s", formatBytes(crc).c_str(),
                     formatBytes(crc_received).c_str());
            valid = false;
            resetAnswerFrame();
            returnCode = ErrorType::WRONG_CRC;
        }
        // breaking after first successfull try
        if (returnCode == ErrorType::DATA_OK) {
            break;
        }
    }

    // ESP_LOGE(TAG, "Returncode: %d", returnCode);
    return returnCode;
}

unsigned int SentecSensorRS485::calculateCRC(byte query[], int length) {
    // Change the last two bytes of the queryFrame to conform to a CRC check
    // Yes, this is necessary. No, I don't know exactly what it does.
    unsigned int tmp1, tmp2, flag;
    tmp1 = 0xFFFF;
    for (unsigned char i = 0; i < length; i++) {
        tmp1 = tmp1 ^ query[i];
        for (unsigned char j = 1; j <= 8; j++) {
            flag = tmp1 & 0x0001;
            tmp1 >>= 1;
            if (flag)
                tmp1 ^= 0xA001;
        }
    }
    // Reverse byte order.
    tmp2 = tmp1 >> 8;
    tmp1 = (tmp1 << 8) | tmp2;
    tmp1 &= 0xFFFF;
    // change last two query bytes
    query[length + 1] = tmp1;
    query[length] = tmp1 >> 8;

    return tmp1; // the returned value is already swapped - CRC_L byte is first & CRC_H byte is last
}

void SentecSensorRS485::printBytes(byte *data, int length) {
    // prints 8-bit data in hex with leading zeroes
    char tmp[16];
    for (int i = 0; i < length; i++) {
        sprintf(tmp, "%.2X", data[i]);
        // sprintf(tmp, "0x%.2X",data[i]);
        Serial.print(tmp);
        Serial.print(" ");
    }
    Serial.println();
}

void SentecSensorRS485::printBytes(word data) {
    char tmp[10];
    sprintf(tmp, "0x%.2X", data >> 8);
    Serial.print(tmp);
    sprintf(tmp, "%.2X", data & 0xFF);
    // sprintf(tmp, "0x%.2X",data[i]);
    Serial.print(tmp);
}

std::string SentecSensorRS485::formatBytes(byte *data, int length) {
    // pls don't hate me for building strings like that.
    // I also wish that I would be good at programming
    std::string tmp;
    // prints 8-bit data in hex with leading zeroes
    char buff[4];
    for (int i = 0; i < length; i++) {
        snprintf(buff, sizeof(buff), "%02X ", data[i]);
        tmp += buff;
    }
    return tmp;
}

std::string SentecSensorRS485::formatBytes(word data) {
    byte arr[2] = {data >> 8, data & 0xFF};
    return formatBytes(arr, 2);
}

word SolarRadiationSensor::getSolarRadiation() {
    readRegister(0, 1);
    glob = word(answerFrame[3], answerFrame[4]);
    ESP_LOGI(TAG, "Global solar radiation: %d W/m^2", glob);
    return glob;
}

String SolarRadiationSensor::getSolarRadiationStr() {
    return getValueStr((int) glob);
}

void RainGaugeSensor::resetPrecipitation() {
    // clears rainfall rata from the rain gauge
    // delay resetting after the last register reading
    delay(100);
    Serial.println("Resetting precipitation sum");
    writeRegister(0x00, 0x5A);
    // TODO check response: matches the sent message exactly
}

void RainGaugeSensor::resetSensor() {
    // no clue what the difference between this one and the previous one is...
    //   and the manual is not helpful, of course
    delay(100);
    Serial.println("Resetting precipitation sum");
    writeRegister(0x37, 0x03);
    // TODO check response: matches the sent message exactly
}

word RainGaugeSensor::getInstantaneousPrecipitation() {
    // gets tips of scale since the last reset, i.e. total precipitation (I THINK)
    //   manual says this is current precipitation - is it?
    readRegister(0, 0x01);
    precipitation = word(answerFrame[3], answerFrame[4]);

    ESP_LOGI(TAG, "Precipitation: %.1f mm", precipitation / 10.0);
    // resetPrecipitation();
    return precipitation;
}

String RainGaugeSensor::getPrecipitationStr() {
    return getValueStr((float) (precipitation / 10.0));
}

float SoilMoistureSensor::getMoistureTemp() {
    readRegister(0, 2); // start register at 0, read 2 variables (vwc, soil temp)
    moistureRaw = (answerFrame[3] << 8) + answerFrame[4];
    // TODO: neg. temp check
    if (answerFrame[5] < 0x80) {
        temperatureRaw = (answerFrame[5] << 8) + answerFrame[6];
    } else {
        temperatureRaw = (answerFrame[5] << 8) + answerFrame[6] - 65536;
    }
    ESP_LOGI(TAG, "Soil moisture:    %.1f  %", (moistureRaw - moistureOffset) / 10.0);
    ESP_LOGI(TAG, "Soil temperature: %.1f °C", (temperatureRaw - temperatureOffset) / 10.0);
    return (temperatureRaw - temperatureOffset) / 10.0;
}

float SoilMoistureSensor::getMoisture() {
    return (moistureRaw - moistureOffset) / 10.0;
}

String SoilMoistureSensor::getMoistureStr() {
    return getValueStr((float) ((moistureRaw - moistureOffset) / 10.0));
}

String SoilMoistureSensor::getTemperatureStr() {
    return getValueStr((float) ((temperatureRaw - temperatureOffset) / 10.0));
}
