#include "dr26.hpp"

static const char *TAG = "DR26";

void ForteDR26::setup() {
    Wire.begin();

    ads.setGain(GAIN_ONE);
    ads.setDataRate(RATE_ADS1115_8SPS);
    if (ads.begin()) {
        ESP_LOGD(TAG, "ADS initialized");
        sensorConnected = true;
    } else {
        ESP_LOGE(TAG, "failed to initialize ADS");
        sensorConnected = false;
    }

    delay(100);
    channel = 0;
}

Measurement ForteDR26::readData() {
    float volts = 0;
    int16_t adc = 0;

    try {
        // TODO: This function might never return if conversionComplete is never true. Check if this is the case
        // FIXME: How to notice from the dr if a value is incorrect?
        adc = ads.readADC_SingleEnded(channel);
        volts = ads.computeVolts(adc);
    }
    catch (NoDataAvailableException &e) {
        // FIXME: This catch block will never be called since we aint throwing any exceptions
        return {ERROR_VALUE, channel, NO_I2C_ADDRESS, MeasurementType::CIRCUMFERENCE_INCREMENT,
                ErrorType::NO_DATA};
    }

    if (!sensorConnected) {
        return {ERROR_VALUE, channel, NO_I2C_ADDRESS, MeasurementType::CIRCUMFERENCE_INCREMENT,
                ErrorType::SENSOR_INIT_FAILED};
    }

    return {volts, channel, NO_I2C_ADDRESS, MeasurementType::CIRCUMFERENCE_INCREMENT,
            ErrorType::DATA_OK};
}

// The following functions change the ADC input range: be careful
// to never to exceed VDD +0.3V max, or to exceed the upper and
// lower limits if you adjust the input range.
// SETTING THE GAIN VALUE INCORRECTLY MAY DESTROY THE ADC!
// The maximum output of the dendrometer is 2.5V, and so a gain of
// one (max 4.096V) or two (max 2.048V) is optimal. Changing the gain
// changes the accuracy of the sensor: higher gain gives a higher
// precision but a smaller range.
//
// GAIN_TWOTHIRDS  // 2/3x gain +/- 6.144V  1 bit = 0.1875mV (default)
// GAIN_ONE        // 1x gain   +/- 4.096V  1 bit = 0.125mV
// GAIN_TWO        // 2x gain   +/- 2.048V  1 bit = 0.0625mV
// GAIN_FOUR       // 4x gain   +/- 1.024V  1 bit = 0.03125mV
// GAIN_EIGHT      // 8x gain   +/- 0.512V  1 bit = 0.015625mV
// GAIN_SIXTEEN    // 16x gain  +/- 0.256V  1 bit = 0.0078125mV
void ForteDR26::changeGain(adsGain_t gain) {
    ads.setGain(gain);
}

void ForteDR26::setChannel(int c) {
    channel = c;
}

std::list<Message> ForteDR26::buildMessages() {
    std::list<Message> messages;
    auto data = readData();
    messages.emplace_back(data,
                          sensorInformation,
                          Time::getInstance().getEpochSeconds());
    return messages;
}

SensorInformation ForteDR26::getSensorInformation() const {
    return sensorInformation;
}


// TODO: What is this?
Adafruit_ADS1115 ForteDR26::ads = ads;

